/* 
 *   $Id: ad_nfs_wrcoll.c,v 1.5 1997/09/28 17:28:57 thakur Exp $    
 *
 *   Copyright (C) 1997 University of Chicago. 
 *   See COPYRIGHT notice in top-level directory.
 */

#include "ad_nfs.h"
#include "adio_extern.h"

/* If ADIOI_GEN_WriteStridedColl is used for NFS, it will
   result in double file locking. Hence the separate code here. It is
   identical to ADIOI_GEN_WriteStridedColl, except that it doesn't call
   ADIO_WriteContig anywhere. It directly calls Unix read/write and locking
   is done across Unix read/write. */ 

/* prototypes of functions used for collective writes only. */
static void ADIOI_Exch_and_write(ADIO_File fd, void *buf, MPI_Datatype
                         datatype, int nprocs, int
                         nprocs_for_coll, int myrank, ADIOI_Access_info
                         *my_req_info, ADIOI_Access_info
                         *others_req_info, ADIO_Offset *offset_list,
                         int *len_list, int contig_access_count, int
                         min_st_offset, ADIO_Offset fd_size,
                         ADIO_Offset *fd_start, ADIO_Offset *fd_end,
                         int *buf_idx);
static void ADIOI_W_Exchange_data(ADIO_File fd, void *buf, char *write_buf,
                         ADIOI_Flatlist_node *flat_buf, ADIO_Offset 
                         *offset_list, int *len_list, int *send_size, 
                         int *recv_size, ADIO_Offset off, int size,
                         int *count, int *curr_offlen_ptr, int
                         *start_pos, int *partial_recv, int
                         *sent_to_proc, int nprocs, 
                         int nprocs_for_coll, int myrank, int
                         buftype_is_contig, int contig_access_count,
                         int min_st_offset, ADIO_Offset fd_size,
                         ADIO_Offset *fd_start, ADIO_Offset *fd_end, 
                         ADIOI_Access_info *others_req_info, 
                         int *send_buf_idx, int *curr_to_proc,
                         int *done_to_proc, int *hole, int iter, 
                         MPI_Aint buftype_extent, int *buf_idx);
static void ADIOI_Fill_send_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node                           *flat_buf, char **send_buf, ADIO_Offset 
                           *offset_list, int *len_list, int *send_size, 
                           MPI_Request *requests, int *sent_to_proc, 
                           int nprocs_for_coll, int myrank, 
                           int contig_access_count, 
                           int min_st_offset, ADIO_Offset fd_size,
                           ADIO_Offset *fd_start, ADIO_Offset *fd_end, 
                           int *send_buf_idx, int *curr_to_proc, 
                           int *done_to_proc, int iter, 
                           MPI_Aint buftype_extent);
static void ADIOI_Heap_merge(ADIOI_Access_info *others_req_info, int *count, 
		      ADIO_Offset *srt_off, int *srt_len, int *start_pos,
		      int nprocs, int nprocs_recv, int total_elements);


void ADIOI_NFS_WriteStridedColl(ADIO_File fd, void *buf, int count,
                       MPI_Datatype datatype, int file_ptr_type,
                       ADIO_Offset offset, ADIO_Status *status, int
                       *error_code)
{
/* Uses a generalized version of the extended two-phase method described
   in "An Extended Two-Phase Method for Accessing Sections of 
   Out-of-Core Arrays", Rajeev Thakur and Alok Choudhary,
   Scientific Programming, (5)4:301--317, Winter 1996. 
   http://www.mcs.anl.gov/home/thakur/ext2ph.ps */

    ADIOI_Access_info *my_req_info; 
    /* array of nprocs access_info structures, one for each other process in
       whose file domain this process's request lies */
    
    ADIOI_Access_info *others_req_info;
    /* array of nprocs access_info structures, one for each other process
       whose request lies in this process's file domain. */

    int i, filetype_is_contig, nprocs, nprocs_for_coll, myrank;
    int *len_list, contig_access_count, interleave_count;
    int buftype_is_contig, size, bufsize, *buf_idx;
    int *count_my_req_per_proc, count_my_req_procs, count_others_req_procs;
    ADIO_Offset *offset_list, start_offset, end_offset, *st_offsets, orig_fp;
    ADIO_Offset *fd_start, *fd_end, fd_size, min_st_offset, *end_offsets;
    ADIO_Offset off;

    ADIOI_Datatype_iscontig(fd->filetype, &filetype_is_contig);
    ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);

    if (filetype_is_contig) {
        /* noncollective is good enough */
        if (buftype_is_contig) {
            MPI_Type_size(datatype, &size);
            bufsize = size * count;
            if (file_ptr_type == ADIO_EXPLICIT_OFFSET) {
		size = fd->etype_size;
                off = fd->disp + size * offset;
                ADIO_WriteContig(fd, buf, bufsize, ADIO_EXPLICIT_OFFSET,
                       off, status, error_code);
            }
            else ADIO_WriteContig(fd, buf, bufsize, ADIO_INDIVIDUAL,
                       0, status, error_code);
        }
        else ADIO_WriteStrided(fd, buf, count, datatype, file_ptr_type,
                       offset, status, error_code);
        return;
    }

/* PFS file pointer modes are not relevant here. */
    if ((fd->iomode != M_ASYNC) && (fd->iomode != M_UNIX)) {
	printf("ADIOI_GEN_WriteStridedColl: only M_ASYNC and M_UNIX iomodes are valid\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

    MPI_Comm_size(fd->comm, &nprocs);
    MPI_Comm_rank(fd->comm, &myrank);

    nprocs_for_coll = ADIOI_MIN(nprocs, ADIOI_MAX_NPROCS_FOR_COLL);
/*  only nprocs_for_coll processes actually perform I/O */

/* For this process's request, calculate the list of offsets and
   lengths in the file and determine the start and end offsets. */

/* Note: end_offset points to the last byte-offset that will be accessed.
         e.g., if start_offset=0 and 100 bytes to be read, end_offset=99*/

    orig_fp = fd->fp_ind;
    ADIOI_Calc_my_off_len(fd, count, datatype, file_ptr_type, offset,
			   &offset_list, &len_list, &start_offset,
			   &end_offset, &contig_access_count); 

/* each process communicates its start and end offsets to other 
   processes. The result is an array each of start and end offsets stored
   in order of process rank. */ 
    
    st_offsets = (ADIO_Offset *) ADIOI_Malloc(nprocs*sizeof(ADIO_Offset));
    end_offsets = (ADIO_Offset *) ADIOI_Malloc(nprocs*sizeof(ADIO_Offset));

    MPI_Allgather(&start_offset, 1, MPI_INT, st_offsets, 1, MPI_INT, 
		  fd->comm);
    MPI_Allgather(&end_offset, 1, MPI_INT, end_offsets, 1, MPI_INT, 
		  fd->comm);
/* MPI_INT because ADIO_Offset is of type MPI_INT */


/* are the accesses of different processes interleaved? */
    interleave_count = 0;
    for (i=1; i<nprocs; i++)
	if (st_offsets[i] < end_offsets[i-1])
	    interleave_count++;
/* This is a rudimentary check for interleaving, but should suffice
   for the moment. */


    if (!interleave_count) {
/* no interleaving of requests. noncollective is good enough */
	ADIOI_Free(offset_list);
	ADIOI_Free(len_list);
	ADIOI_Free(st_offsets);
	ADIOI_Free(end_offsets);	

	fd->fp_ind = orig_fp;
	ADIO_WriteStrided(fd, buf, count, datatype, file_ptr_type,
                          offset, status, error_code);
	return;
    }

/* Divide the I/O workload among "nprocs_for_coll" processes. This is
   done by (logically) dividing the file into file domains (FDs); each
   process may directly access only its own file domain. */

    ADIOI_Calc_file_domains(st_offsets, end_offsets, nprocs,
			    nprocs_for_coll, &min_st_offset,
			    &fd_start, &fd_end, &fd_size);   


/* calculate what portions of the access requests of this process are
   located in the file domains of other processes */

    ADIOI_Calc_my_req_info(offset_list, len_list, contig_access_count,
			   min_st_offset, fd_start, fd_end, fd_size,
			   nprocs, nprocs_for_coll, &count_my_req_procs, 
			   &count_my_req_per_proc, &my_req_info,
                           &buf_idx); 

/* based on everyone's my_req_info, calculate what requests of other
   processes lie in this process's file domain.
   count_others_req_procs = number of processes whose requests lie in
   this process's file domain (including this process itself) 
   count_others_req_per_proc[i] indicates how many separate contiguous
   requests of proc. i lie in this process's file domain. */

    ADIOI_Calc_others_req_info(fd, count_my_req_procs, 
			       count_my_req_per_proc, my_req_info, 
			       nprocs, myrank, nprocs_for_coll,
			       &count_others_req_procs, &others_req_info); 

    ADIOI_Free(count_my_req_per_proc);

/* exchange data and write in sizes of no more than ADIOI_COLL_BUFSIZE. */
    ADIOI_Exch_and_write(fd, buf, datatype, nprocs, nprocs_for_coll, myrank,
                        my_req_info, others_req_info, offset_list,
			len_list, contig_access_count, min_st_offset,
			fd_size, fd_start, fd_end, buf_idx);

    if (!buftype_is_contig) ADIOI_Delete_flattened(datatype);

/* free all memory allocated for collective I/O */

    for (i=0; i<nprocs_for_coll; i++) {
	if (my_req_info[i].count) {
	    ADIOI_Free(my_req_info[i].offsets);
	    ADIOI_Free(my_req_info[i].lens);
	}
    }
    ADIOI_Free(my_req_info);

    for (i=0; i<nprocs; i++) {
	if (others_req_info[i].count) {
	    ADIOI_Free(others_req_info[i].offsets);
	    ADIOI_Free(others_req_info[i].lens);
	    ADIOI_Free(others_req_info[i].mem_ptrs);
	}
    }
    ADIOI_Free(others_req_info);

    ADIOI_Free(buf_idx);
    ADIOI_Free(offset_list);
    ADIOI_Free(len_list);
    ADIOI_Free(st_offsets);
    ADIOI_Free(end_offsets);
    ADIOI_Free(fd_start);
    ADIOI_Free(fd_end);

    fd->fp_sys_posn = -1;   /* set it to null. */

    *error_code = MPI_SUCCESS;
}



static void ADIOI_Exch_and_write(ADIO_File fd, void *buf, MPI_Datatype
			 datatype, int nprocs, int
			 nprocs_for_coll, int myrank, ADIOI_Access_info
			 *my_req_info, ADIOI_Access_info
			 *others_req_info, ADIO_Offset *offset_list,
			 int *len_list, int contig_access_count, int
                         min_st_offset, ADIO_Offset fd_size,
			 ADIO_Offset *fd_start, ADIO_Offset *fd_end,
                         int *buf_idx)
{
/* Send data to appropriate processes and write in sizes of no more
   than ADIOI_COLL_BUFSIZE.    
   The idea is to reduce the amount of extra memory required for
   collective I/O. If all data were written all at once, which is much
   easier, it would require temp space more than the size of user_buf,
   which is often unacceptable. For example, to write a distributed
   array to a file, where each local array is 8Mbytes, requiring
   at least another 8Mbytes of temp space is unacceptable. */

    int hole, i, j, m, size, ntimes, max_ntimes, buftype_is_contig;
    ADIO_Offset st_loc=0, end_loc=0, off, done, req_off;
    char *write_buf;
    int *curr_offlen_ptr, *count, *send_size, req_len, *recv_size;
    int *partial_recv, *sent_to_proc, *start_pos, err, flag;
    int *send_buf_idx, *curr_to_proc, *done_to_proc;
    ADIOI_Flatlist_node *flat_buf;
    MPI_Aint buftype_extent;

/* calculate the number of writes of size ADIOI_COLL_BUFSIZE
   to be done by each process and the max among all processes.
   That gives the no. of communication phases as well. */

    for (i=0; i<nprocs; i++) {
	if (others_req_info[i].count) {
	    st_loc = others_req_info[i].offsets[0];
	    end_loc = others_req_info[i].offsets[0];
	    break;
	}
    }

    for (i=0; i<nprocs; i++)
	for (j=0; j<others_req_info[i].count; j++) {
	    st_loc = ADIOI_MIN(st_loc, others_req_info[i].offsets[j]);
	    end_loc = ADIOI_MAX(end_loc, (others_req_info[i].offsets[j]
				       + others_req_info[i].lens[j] - 1));
	}

/* ntimes=ceiling_div(end_loc - st_loc + 1, ADIOI_COLL_BUFSIZE)*/

    ntimes = (end_loc - st_loc + ADIOI_COLL_BUFSIZE)/ADIOI_COLL_BUFSIZE;

    if (!st_loc && !end_loc) ntimes = 0; /* this process does no I/O. */

    MPI_Allreduce(&ntimes, &max_ntimes, 1, MPI_INT, MPI_MAX,
		  fd->comm); 

    if (ntimes) write_buf = (char *) ADIOI_Malloc(ADIOI_COLL_BUFSIZE);

    curr_offlen_ptr = (int *) ADIOI_Calloc(nprocs, sizeof(int)); 
    /* its use is explained below. calloc initializes to 0. */

    count = (int *) ADIOI_Malloc(nprocs*sizeof(int));
    /* to store count of how many off-len pairs per proc are satisfied
       in an iteration. */

    partial_recv = (int *) ADIOI_Calloc(nprocs, sizeof(int));
    /* if only a portion of the last off-len pair is recd. from a process
       in a particular iteration, the length recd. is stored here.
       calloc initializes to 0. */

    send_size = (int *) ADIOI_Malloc(nprocs*sizeof(int));
    /* total size of data to be sent to each proc. in an iteration.
       Of size nprocs so that I can use MPI_Alltoall later. */

    recv_size = (int *) ADIOI_Malloc(nprocs*sizeof(int));
    /* total size of data to be recd. from each proc. in an iteration.*/

    sent_to_proc = (int *) ADIOI_Calloc(nprocs_for_coll, sizeof(int));
    /* amount of data sent to each proc so far. Used in
       ADIOI_Fill_send_buffer. initialized to 0 here. */

    send_buf_idx = (int *) ADIOI_Malloc(nprocs_for_coll*sizeof(int));
    curr_to_proc = (int *) ADIOI_Malloc(nprocs_for_coll*sizeof(int));
    done_to_proc = (int *) ADIOI_Malloc(nprocs_for_coll*sizeof(int));
    /* Above three are used in ADIOI_Fill_send_buffer*/

    start_pos = (int *) ADIOI_Malloc(nprocs*sizeof(int));
    /* used to store the starting value of curr_offlen_ptr[i] in 
       this iteration */

    ADIOI_Datatype_iscontig(datatype, &buftype_is_contig);
    if (!buftype_is_contig) {
	ADIOI_Flatten_datatype(datatype);
	flat_buf = ADIOI_Flatlist;
        while (flat_buf->type != datatype) flat_buf = flat_buf->next;
    }
    MPI_Type_extent(datatype, &buftype_extent);


/* I need to check if there are any outstanding nonblocking writes to
   the file, which could potentially interfere with the writes taking
   place in this collective write call. Since this is not likely to be
   common, let me do the simplest thing possible here: Each process
   completes all pending nonblocking operations before completing. */

    ADIOI_Complete_async(&err);
    MPI_Barrier(fd->comm);

    done = 0;
    off = st_loc;

    for (m=0; m<ntimes; m++) {
       /* go through all others_req_info and check which will be satisfied
          by the current write */

       /* Note that MPI guarantees that displacements in filetypes are in 
          monotonically nondecreasing order and that, for writes, the
	  filetypes cannot specify overlapping regions in the file. This
	  simplifies implementation a bit compared to reads. */

          /* off = start offset in the file for the data to be written in 
                   this iteration 
             size = size of data written corresponding to off
             req_off = off in file for a particular contiguous request 
                       minus what was satisfied in previous iteration
             req_size = size corresponding to req_off */

	/* first calculate what should be communicated */

	for (i=0; i<nprocs; i++) count[i] = recv_size[i] = 0;

	size = ADIOI_MIN(ADIOI_COLL_BUFSIZE, end_loc-st_loc+1-done); 

	for (i=0; i<nprocs; i++) {
	    if (others_req_info[i].count) {
		start_pos[i] = curr_offlen_ptr[i];
		for (j=curr_offlen_ptr[i]; j<others_req_info[i].count;
		     j++) {
		    if (partial_recv[i]) {
			/* this request may have been partially
			   satisfied in the previous iteration. */
			req_off = others_req_info[i].offsets[j] +
			    partial_recv[i]; 
                        req_len = others_req_info[i].lens[j] -
			    partial_recv[i];
			partial_recv[i] = 0;
			/* modify the off-len pair to reflect this change */
			others_req_info[i].offsets[j] = req_off;
			others_req_info[i].lens[j] = req_len;
		    }
		    else {
			req_off = others_req_info[i].offsets[j];
                        req_len = others_req_info[i].lens[j];
		    }
		    if (req_off < off + size) {
			count[i]++;
			MPI_Address(write_buf+req_off-off, 
                               &(others_req_info[i].mem_ptrs[j]));
			recv_size[i] += ADIOI_MIN(off + size - 
						  req_off, req_len);

			if (off+size-req_off < req_len) {
			    partial_recv[i] = off + size - req_off;
			    if ((j+1 < others_req_info[i].count) && 
                                 (others_req_info[i].offsets[j+1] < 
                                     off+size)) { 
				printf("Error: the filetype specifies an overlapping write\n");
				MPI_Abort(MPI_COMM_WORLD, 1);
			    }
			    break;
			}
		    }
		    else break;
		}
		curr_offlen_ptr[i] = j;
	    }
	}
	
	ADIOI_W_Exchange_data(fd, buf, write_buf, flat_buf, offset_list, 
                            len_list, send_size, recv_size, off, size, count, 
                            curr_offlen_ptr, start_pos, partial_recv, 
                            sent_to_proc, nprocs, nprocs_for_coll, myrank, 
			    buftype_is_contig, contig_access_count,
			    min_st_offset, fd_size, fd_start, fd_end,
			    others_req_info, send_buf_idx, curr_to_proc,
                            done_to_proc, &hole, m, buftype_extent, buf_idx); 
	
	flag = 0;
	for (i=0; i<nprocs; i++)
	    if (count[i]) flag = 1;

	if (flag) {
	    lseek(fd->fd_sys, off, SEEK_SET);
	    if (!hole || (fd->access_mode & ADIO_UNIQUE_OPEN)) {
		/* The file is not open elsewhere or there was no
		   hole. I need to do the usual lock for NFS and
		   release it immediately. */ 
		ADIOI_WRITE_LOCK(fd->fd_sys, off, SEEK_SET, size);
		err = write(fd->fd_sys, write_buf, size);
		ADIOI_UNLOCK(fd->fd_sys, off, SEEK_SET, size);
	    }
	    else {
		/* The file may be open elsewhere and there was a
		   hole, so the file segment was locked in
		   ADIOI_W_Exchange_data. */ 
		err = write(fd->fd_sys, write_buf, size);
		ADIOI_UNLOCK(fd->fd_sys, off, SEEK_SET, size);
	    }
	}

	off += size;
	done += size;
    }

    for (i=0; i<nprocs; i++) count[i] = recv_size[i] = 0;
    for (m=ntimes; m<max_ntimes; m++) 
/* nothing to recv, but check for send. */
	ADIOI_W_Exchange_data(fd, buf, write_buf, flat_buf, offset_list, 
                            len_list, send_size, recv_size, off, size, count, 
                            curr_offlen_ptr, start_pos, partial_recv, 
                            sent_to_proc, nprocs, nprocs_for_coll, myrank, 
			    buftype_is_contig, contig_access_count,
			    min_st_offset, fd_size, fd_start, fd_end,
			    others_req_info, send_buf_idx, 
                            curr_to_proc, done_to_proc, &hole, m, 
                            buftype_extent, buf_idx); 

    if (ntimes) ADIOI_Free(write_buf);
    ADIOI_Free(curr_offlen_ptr);
    ADIOI_Free(count);
    ADIOI_Free(partial_recv);
    ADIOI_Free(send_size);
    ADIOI_Free(recv_size);
    ADIOI_Free(sent_to_proc);
    ADIOI_Free(start_pos);
    ADIOI_Free(send_buf_idx);
    ADIOI_Free(curr_to_proc);
    ADIOI_Free(done_to_proc);
}



static void ADIOI_W_Exchange_data(ADIO_File fd, void *buf, char *write_buf,
                         ADIOI_Flatlist_node *flat_buf, ADIO_Offset 
                         *offset_list, int *len_list, int *send_size, 
                         int *recv_size, ADIO_Offset off, int size,
                         int *count, int *curr_offlen_ptr, int
                         *start_pos, int *partial_recv, int
                         *sent_to_proc, int nprocs, 
                         int nprocs_for_coll, int myrank, int
                         buftype_is_contig, int contig_access_count,
                         int min_st_offset, ADIO_Offset fd_size,
                         ADIO_Offset *fd_start, ADIO_Offset *fd_end, 
                         ADIOI_Access_info *others_req_info, 
                         int *send_buf_idx, int *curr_to_proc,
                         int *done_to_proc, int *hole, int iter, 
                         MPI_Aint buftype_extent, int *buf_idx)
{
    int i, j, k, *tmp_len, nprocs_recv, nprocs_send, err;
    char **send_buf; 
    MPI_Request *requests;
    MPI_Datatype *recv_types;
    MPI_Status *statuses;
    int *srt_len, sum;
    ADIO_Offset *srt_off;

/* exchange recv_size info so that each process knows how much to
   send to whom. */

    MPI_Alltoall(recv_size, 1, MPI_INT, send_size, 1, MPI_INT, fd->comm);

    /* create derived datatypes for recv */

    nprocs_recv = 0;
    for (i=0; i<nprocs; i++) if (recv_size[i]) nprocs_recv++;

    recv_types = (MPI_Datatype *)
	ADIOI_Malloc((nprocs_recv+1)*sizeof(MPI_Datatype)); 
/* +1 to avoid a 0-size malloc */

    tmp_len = (int *) ADIOI_Malloc(nprocs*sizeof(int));
    j = 0;
    for (i=0; i<nprocs; i++) {
	if (recv_size[i]) {
/* take care if the last off-len pair is a partial recv */
	    if (partial_recv[i]) {
		k = start_pos[i] + count[i] - 1;
		tmp_len[i] = others_req_info[i].lens[k];
		others_req_info[i].lens[k] = partial_recv[i];
	    }
	    MPI_Type_hindexed(count[i], 
                 &(others_req_info[i].lens[start_pos[i]]),
	             &(others_req_info[i].mem_ptrs[start_pos[i]]), 
			 MPI_BYTE, recv_types+j);
	    /* absolute displacements; use MPI_BOTTOM in recv */
	    MPI_Type_commit(recv_types+j);
	    j++;
	}
    }

    /* To avoid a read-modify-write, check if there are holes in the 
       data to be written. For this, merge the (sorted) offset lists
       others_req_info using a heap-merge. */

    sum = 0;
    for (i=0; i<nprocs; i++) sum += count[i];
    srt_off = (ADIO_Offset *) ADIOI_Malloc((sum+1)*sizeof(ADIO_Offset));
    srt_len = (int *) ADIOI_Malloc((sum+1)*sizeof(int));
    /* +1 to avoid a 0-size malloc */

    ADIOI_Heap_merge(others_req_info, count, srt_off, srt_len, start_pos,
		     nprocs, nprocs_recv, sum);

/*    if (myrank == 0) {
	printf("\n\n SORTED: \n");
	for (i=0; i<sum; i++)
	    printf("%ld ", srt_off[i]);
	printf("\n\n");
	for (i=0; i<sum; i++)
	    printf("%d ", srt_len[i]);
    }*/

/* for partial recvs, restore original lengths */
    for (i=0; i<nprocs; i++) 
	if (partial_recv[i]) {
	    k = start_pos[i] + count[i] - 1;
	    others_req_info[i].lens[k] = tmp_len[i];
	}
    ADIOI_Free(tmp_len);

/* check if there are any holes */
    *hole = 0;
    for (i=0; i<sum-1; i++)
        if (srt_off[i]+srt_len[i] < srt_off[i+1]) {
            *hole = 1;
            break;
        }

    ADIOI_Free(srt_off);
    ADIOI_Free(srt_len);

    if (nprocs_recv) {
        lseek(fd->fd_sys, off, SEEK_SET);
	if (*hole) {
	    if (fd->access_mode & ADIO_UNIQUE_OPEN) {
		/* The file is not open elsewhere, so I only need to do
		   the usual lock for NFS and release it immediately. */
		ADIOI_READ_LOCK(fd->fd_sys, off, SEEK_SET, size);
		err = read(fd->fd_sys, write_buf, size);
		ADIOI_UNLOCK(fd->fd_sys, off, SEEK_SET, size);
	    }
	    else {
		/* The file may be opened elsewhere, so I need to lock the
		   file segment corresponding to (off,size) and then do a
		   read-modify-write. */ 
		ADIOI_WRITE_LOCK(fd->fd_sys, off, SEEK_SET, size);
		err = read(fd->fd_sys, write_buf, size);
	    }
	}
    }

    nprocs_send = 0;
    for (i=0; i<nprocs_for_coll; i++) if (send_size[i]) nprocs_send++;

    requests = (MPI_Request *) 
        ADIOI_Malloc((nprocs_send+nprocs_recv+1)*sizeof(MPI_Request)); 
/* +1 to avoid a 0-size malloc */

/* post receives */
    j = 0;
    for (i=0; i<nprocs; i++) {
        if (recv_size[i]) {
            MPI_Irecv(MPI_BOTTOM, 1, recv_types[j], i, myrank+i+100*iter,
                      fd->comm, requests+j);
            j++;
        }
    }

/* post sends. if buftype_is_contig, data can be directly sent from
   user buf at location given by buf_idx. else use send_buf. */

    if (buftype_is_contig) {
        j = 0;
        for (i=0; i<nprocs_for_coll; i++) 
            if (send_size[i]) {
                MPI_Isend(((char *) buf) + buf_idx[i], send_size[i], 
                            MPI_BYTE, i,  myrank+i+100*iter, fd->comm, 
                                  requests+nprocs_recv+j);
                j++;
                buf_idx[i] += send_size[i];
            }
    }
    else if (nprocs_send) {
        /* buftype is not contig */
        send_buf = (char **) ADIOI_Malloc(nprocs_for_coll*sizeof(char*));
        for (i=0; i<nprocs_for_coll; i++) 
            if (send_size[i]) 
                send_buf[i] = (char *) ADIOI_Malloc(send_size[i]);

        ADIOI_Fill_send_buffer(fd, buf, flat_buf, send_buf,
                           offset_list, len_list, send_size, 
                           requests+nprocs_recv,
                           sent_to_proc, nprocs_for_coll, myrank, 
                           contig_access_count,
                           min_st_offset, fd_size, fd_start, fd_end, 
                           send_buf_idx, curr_to_proc, done_to_proc, iter,
                           buftype_extent);
        /* the send is done in ADIOI_Fill_send_buffer */
    }

    for (i=0; i<nprocs_recv; i++) MPI_Type_free(recv_types+i);
    ADIOI_Free(recv_types);
    
    statuses = (MPI_Status *) ADIOI_Malloc((nprocs_send+nprocs_recv+1) * \
                                     sizeof(MPI_Status)); 
     /* +1 to avoid a 0-size malloc */

#ifdef __NEEDS_MPI_TEST
    i = 0;
    while (!i) MPI_Testall(nprocs_send+nprocs_recv, requests, &i, statuses);
#else
    MPI_Waitall(nprocs_send+nprocs_recv, requests, statuses);
#endif

    ADIOI_Free(statuses);
    ADIOI_Free(requests);
    if (!buftype_is_contig && nprocs_send) {
        for (i=0; i<nprocs_for_coll; i++) 
            if (send_size[i]) ADIOI_Free(send_buf[i]);
        ADIOI_Free(send_buf);
    }
}


#define ADIOI_BUF_INCR \
{ \
    while (buf_incr) { \
        size_in_buf = ADIOI_MIN(buf_incr, flat_buf_sz); \
        user_buf_idx += size_in_buf; \
        flat_buf_sz -= size_in_buf; \
        if (!flat_buf_sz) { \
            if (flat_buf_idx < (flat_buf->count - 1)) flat_buf_idx++; \
            else { \
                flat_buf_idx = 0; \
                n_buftypes++; \
            } \
            user_buf_idx = flat_buf->indices[flat_buf_idx] + \
                              n_buftypes*buftype_extent; \
            flat_buf_sz = flat_buf->blocklens[flat_buf_idx]; \
        } \
        buf_incr -= size_in_buf; \
    } \
}


#define ADIOI_BUF_COPY \
{ \
    while (size) { \
        size_in_buf = ADIOI_MIN(size, flat_buf_sz); \
        memcpy(&(send_buf[p][send_buf_idx[p]]), \
               ((char *) buf) + user_buf_idx, size_in_buf); \
        send_buf_idx[p] += size_in_buf; \
        user_buf_idx += size_in_buf; \
        flat_buf_sz -= size_in_buf; \
        if (!flat_buf_sz) { \
            if (flat_buf_idx < (flat_buf->count - 1)) flat_buf_idx++; \
            else { \
                flat_buf_idx = 0; \
                n_buftypes++; \
            } \
            user_buf_idx = flat_buf->indices[flat_buf_idx] + \
                              n_buftypes*buftype_extent; \
            flat_buf_sz = flat_buf->blocklens[flat_buf_idx]; \
        } \
        size -= size_in_buf; \
        buf_incr -= size_in_buf; \
    } \
    ADIOI_BUF_INCR \
}


static void ADIOI_Fill_send_buffer(ADIO_File fd, void *buf, ADIOI_Flatlist_node
                           *flat_buf, char **send_buf, ADIO_Offset 
                           *offset_list, int *len_list, int *send_size, 
                           MPI_Request *requests, int *sent_to_proc, 
                           int nprocs_for_coll, int myrank, 
                           int contig_access_count, 
                           int min_st_offset, ADIO_Offset fd_size,
                           ADIO_Offset *fd_start, ADIO_Offset *fd_end, 
                           int *send_buf_idx, int *curr_to_proc, 
                           int *done_to_proc, int iter,
                           MPI_Aint buftype_extent)
{
/* this function is only called if buftype is not contig */

    int i, p, flat_buf_idx, user_buf_idx, rem_len, size;
    int flat_buf_sz, len, buf_incr, size_in_buf, j, jj, n_buftypes;
    ADIO_Offset off;

/*  curr_to_proc[p] = amount of data sent to proc. p that has already
    been accounted for so far
    done_to_proc[p] = amount of data already sent to proc. p in 
    previous iterations
    user_buf_idx = current location in user buffer 
    send_buf_idx[p] = current location in send_buf of proc. p  */

    for (i=0; i<nprocs_for_coll; i++) {
	send_buf_idx[i] = curr_to_proc[i] = 0;
	done_to_proc[i] = sent_to_proc[i];
    }
    jj = 0;

    user_buf_idx = flat_buf->indices[0];
    flat_buf_idx = 0;
    n_buftypes = 0;
    flat_buf_sz = flat_buf->blocklens[0];

    /* flat_buf_idx = current index into flattened buftype
       flat_buf_sz = size of current contiguous component in 
	                 flattened buf */

    for (i=0; i<contig_access_count; i++) { 
	p = (offset_list[i]-min_st_offset+fd_size)/fd_size - 1;
	off = offset_list[i];
	len = ((off+len_list[i]-1) <= fd_end[p]) ? 
		len_list[i] : (fd_end[p] - off + 1);

	if (send_buf_idx[p] < send_size[p]) {
	    if (curr_to_proc[p]+len > done_to_proc[p]) {
		if (done_to_proc[p] > curr_to_proc[p]) {
		    size = ADIOI_MIN(curr_to_proc[p]+len-done_to_proc[p], 
                                         send_size[p]-send_buf_idx[p]);
		    buf_incr = done_to_proc[p] - curr_to_proc[p];
		    ADIOI_BUF_INCR
		    buf_incr = curr_to_proc[p] + len - done_to_proc[p];
		    curr_to_proc[p] = done_to_proc[p] + size;
		    ADIOI_BUF_COPY
		}
		else {
		    size = ADIOI_MIN(len, send_size[p]-send_buf_idx[p]); 
		    buf_incr = len;
		    curr_to_proc[p] += size;
		    ADIOI_BUF_COPY
		}
		if (send_buf_idx[p] == send_size[p]) {
		    MPI_Isend(send_buf[p], send_size[p], MPI_BYTE, p, 
                                myrank+p+100*iter, fd->comm, requests+jj);
		    jj++;
		}
	    }
	    else {
		curr_to_proc[p] += len;
		buf_incr = len;
		ADIOI_BUF_INCR
	    }
	}
	else {
	    buf_incr = len;
	    ADIOI_BUF_INCR
	}
            
	/*this request may span the file domains of more than one process*/
	rem_len = len_list[i] - len;
	while (rem_len != 0) {
	    p++;
	    off = fd_start[p];
	    len = ((off+rem_len-1) <= fd_end[p]) ? rem_len : 
		    (fd_end[p] - off + 1);
	    if (send_buf_idx[p] < send_size[p]) {
		if (curr_to_proc[p]+len > done_to_proc[p]) {
		    if (done_to_proc[p] > curr_to_proc[p]) {
			size = ADIOI_MIN(curr_to_proc[p] + len - 
                                done_to_proc[p], send_size[p]-send_buf_idx[p]);
			buf_incr = done_to_proc[p] - curr_to_proc[p];
			ADIOI_BUF_INCR
		        buf_incr = curr_to_proc[p] + len - done_to_proc[p];
			curr_to_proc[p] = done_to_proc[p] + size;
		        ADIOI_BUF_COPY
		    }
		    else {
			size = ADIOI_MIN(len,send_size[p]-send_buf_idx[p]);
			buf_incr = len;
			curr_to_proc[p] += size;
			ADIOI_BUF_COPY
		    }
		    if (send_buf_idx[p] == send_size[p]) {
			MPI_Isend(send_buf[p], send_size[p], MPI_BYTE, p, 
				myrank+p+100*iter, fd->comm, requests+jj);
			jj++;
		    }
		}
		else {
		    curr_to_proc[p] += len;
		    buf_incr = len;
		    ADIOI_BUF_INCR
		}
	    }
	    else {
		buf_incr = len;
		ADIOI_BUF_INCR
            }
	    rem_len -= len;
	}
    }
    for (i=0; i<nprocs_for_coll; i++) 
	if (send_size[i]) sent_to_proc[i] = curr_to_proc[i];
}


static void ADIOI_Heap_merge(ADIOI_Access_info *others_req_info, int *count, 
		      ADIO_Offset *srt_off, int *srt_len, int *start_pos,
		      int nprocs, int nprocs_recv, int total_elements)
{
    typedef struct {
	ADIO_Offset *off_list;
	int *len_list;
	int nelem;
    } heap_struct;

    heap_struct *a, tmp;
    int i, j, heapsize, l, r, k, smallest;

    a = (heap_struct *) ADIOI_Malloc((nprocs_recv+1)*sizeof(heap_struct));

    j = 0;
    for (i=0; i<nprocs; i++)
	if (count[i]) {
	    a[j].off_list = &(others_req_info[i].offsets[start_pos[i]]);
	    a[j].len_list = &(others_req_info[i].lens[start_pos[i]]);
	    a[j].nelem = count[i];
	    j++;
	}

    /* build a heap out of the first element from each list, with
       the smallest element of the heap at the root */

    heapsize = nprocs_recv;
    for (i=heapsize/2 - 1; i>=0; i--) {
	/* Heapify(a, i, heapsize); Algorithm from Cormen et al. pg. 143
           modified for a heap with smallest element at root. I have 
           removed the recursion so that there are no function calls.
           Function calls are too expensive. */
	k = i;
	while (1) {
	    l = 2*(k+1) - 1;
	    r = 2*(k+1);

	    if ((l < heapsize) && 
		(*(a[l].off_list) < *(a[k].off_list)))
		smallest = l;
	    else smallest = k;

	    if ((r < heapsize) && 
		(*(a[r].off_list) < *(a[smallest].off_list)))
		smallest = r;

	    if (smallest != k) {
		tmp.off_list = a[k].off_list;
		tmp.len_list = a[k].len_list;
		tmp.nelem = a[k].nelem;

		a[k].off_list = a[smallest].off_list;
		a[k].len_list = a[smallest].len_list;
		a[k].nelem = a[smallest].nelem;
		
		a[smallest].off_list = tmp.off_list;
		a[smallest].len_list = tmp.len_list;
		a[smallest].nelem = tmp.nelem;
	    
		k = smallest;
	    }
	    else break;
	}
    }

    for (i=0; i<total_elements; i++) {
        /* extract smallest element from heap, i.e. the root */
	srt_off[i] = *(a[0].off_list);
	srt_len[i] = *(a[0].len_list);
	(a[0].nelem)--;

	if (!a[0].nelem) {
	    a[0].off_list = a[heapsize-1].off_list;
	    a[0].len_list = a[heapsize-1].len_list;
	    a[0].nelem = a[heapsize-1].nelem;
	    heapsize--;
	}
	else {
	    (a[0].off_list)++;
	    (a[0].len_list)++;
	}

	/* Heapify(a, 0, heapsize); */
	k = 0;
	while (1) {
	    l = 2*(k+1) - 1;
	    r = 2*(k+1);

	    if ((l < heapsize) && 
		(*(a[l].off_list) < *(a[k].off_list)))
		smallest = l;
	    else smallest = k;

	    if ((r < heapsize) && 
		(*(a[r].off_list) < *(a[smallest].off_list)))
		smallest = r;

	    if (smallest != k) {
		tmp.off_list = a[k].off_list;
		tmp.len_list = a[k].len_list;
		tmp.nelem = a[k].nelem;

		a[k].off_list = a[smallest].off_list;
		a[k].len_list = a[smallest].len_list;
		a[k].nelem = a[smallest].nelem;
		
		a[smallest].off_list = tmp.off_list;
		a[smallest].len_list = tmp.len_list;
		a[smallest].nelem = tmp.nelem;
	    
		k = smallest;
	    }
	    else break;
	}
    }
    ADIOI_Free(a);
}
