/* 
 *   $Id: open.c,v 1.9 1997/10/01 07:56:36 thakur Exp $    
 *
 *   Copyright (C) 1997 University of Chicago. 
 *   See COPYRIGHT notice in top-level directory.
 */

#include "mpioimpl.h"

#ifndef __MPIO_BUILD_PROFILING
int ADIO_Init_keyval=MPI_KEYVAL_INVALID;
#else
extern int ADIO_Init_keyval;
#endif

/*@
    MPI_File_open - Opens a file

Input Parameters:
. comm - communicator (handle)
. filename - name of file to open (string)
. amode - file access mode (integer)
. info - info object (handle)

Output Parameters:
. fh - file handle (handle)

.N fortran
@*/
int MPI_File_open(MPI_Comm comm, char *filename, int amode, 
                  MPI_Info info, MPI_File *fh)
{
    int error_code, file_system, flag; 

    if (comm == MPI_COMM_NULL) {
	printf("MPI_File_open: Invalid communicator\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

    if (!(amode & MPI_MODE_RDONLY) && !(amode & MPI_MODE_RDWR)
	  && !(amode & MPI_MODE_WRONLY) && !(amode & MPI_MODE_CREATE) 
            && !(amode & MPI_MODE_UNIQUE_OPEN) 
              && !(amode & MPI_MODE_DELETE_ON_CLOSE) 
                && !(amode & MPI_MODE_EXCL) 
                  && !(amode & MPI_MODE_APPEND)) {
	printf("MPI_File_open: Invalid amode\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

/* check if ADIO has been initialized. If not, initialize it */
    if (ADIO_Init_keyval == MPI_KEYVAL_INVALID) {

/* check if MPI itself has been initialized. If not, flag an error.
   Can't initialize it here, because don't know argc, argv */
	MPI_Initialized(&flag);
	if (!flag) {
	    printf("Error: MPI_Init() must be called before using MPI-IO\n");
	    MPI_Abort(MPI_COMM_WORLD, 1);
	}

	MPI_Keyval_create(MPI_NULL_COPY_FN, ADIOI_End_call, &ADIO_Init_keyval,
			  (void *) 0);  

/* put a dummy attribute on MPI_COMM_WORLD, because we want the delete
   function to be called when MPI_COMM_WORLD is freed. Hopefully the
   MPI library frees MPI_COMM_WORLD when MPI_Finalize is called,
   though the standard does not mandate this. */

	MPI_Attr_put(MPI_COMM_WORLD, ADIO_Init_keyval, (void *) 0);

/* initialize ADIO */

	ADIO_Init( (int *)0, (char ***)0, &error_code);
    }

#ifndef __PFS
    if (!strncmp(filename, "pfs:", 4) || !strncmp(filename, "PFS:", 4)) {
	printf("MPI_File_open: ROMIO has not been configured to use the PFS file system\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
#endif
#ifndef __PIOFS
    if (!strncmp(filename, "piofs:", 6) || !strncmp(filename, "PIOFS:", 6)) {
	printf("MPI_File_open: ROMIO has not been configured to use the PIOFS file system\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
#endif
#ifndef __UFS
    if (!strncmp(filename, "ufs:", 4) || !strncmp(filename, "UFS:", 4)) {
	printf("MPI_File_open: ROMIO has not been configured to use the UFS file system\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
#endif
#ifndef __NFS
    if (!strncmp(filename, "nfs:", 4) || !strncmp(filename, "NFS:", 4)) {
	printf("MPI_File_open: ROMIO has not been configured to use the NFS file system\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }
#endif

    if (!strncmp(filename, "pfs:", 4) || !strncmp(filename, "PFS:", 4)) {
	file_system = ADIO_PFS;
	filename += 4;
    }
    else if (!strncmp(filename, "piofs:", 6) || !strncmp(filename, "PIOFS:", 6)) {
	file_system = ADIO_PIOFS;
	filename += 6;
    }
    else if (!strncmp(filename, "ufs:", 4) || !strncmp(filename, "UFS:", 4)) {
	file_system = ADIO_UFS;
	filename += 4;
    }
    else if (!strncmp(filename, "nfs:", 4) || !strncmp(filename, "NFS:", 4)) {
	file_system = ADIO_NFS;
	filename += 4;
    }
    else {
        printf("MPI_File_open: The filename must be prefixed with a string to indicate the type of file sytem (piofs:, pfs:, nfs:, ufs:).\n");
	MPI_Abort(MPI_COMM_WORLD, 1);
    }

/* use default values for disp, etype, filetype */    
/* set iomode=M_ASYNC. It is used to implement the Intel PFS interface
   on top of ADIO. Not relevant for MPI-IO implementation */    

    *fh = ADIO_Open(comm, filename, file_system, amode, 0, MPI_BYTE,
                     MPI_BYTE, M_ASYNC, info, ADIO_PERM_NULL, &error_code);

    return error_code;
}



/* This is the delete callback function associated with
   ADIO_Init_keyval when MPI_COMM_WORLD is freed */

#ifndef __MPIO_BUILD_PROFILING
int ADIOI_End_call(MPI_Comm comm, int keyval, void *attribute_val, void
		  *extra_state)
{
    int error_code;

    ADIO_End(&error_code);
    return error_code;
}
#endif
